(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "./util", "./constants", "./pack-writer"], factory);
    }
})(function (require, exports) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    const util_1 = require("./util");
    const constants_1 = require("./constants");
    const pack_writer_1 = require("./pack-writer");
    const arrayItemWriters = [
        new pack_writer_1.ZipsonStringPackWriter(),
    ];
    function isObjectInitialTempletable(a, b, struct, level = 0) {
        if (level > 3) {
            return false;
        }
        const keysA = Object.keys(a);
        const keysB = Object.keys(b);
        let hasNested = false;
        if (keysA.length !== keysB.length) {
            return false;
        }
        if (keysA.length > 10) {
            return false;
        }
        keysA.sort((a, b) => a.localeCompare(b));
        keysB.sort((a, b) => a.localeCompare(b));
        for (let i = 0; i < keysA.length; i++) {
            if (keysA[i] !== keysB[i]) {
                return false;
            }
            if (isObject(a[keysA[i]])) {
                if (!isObject(b[keysB[i]])) {
                    return false;
                }
                const nextStruct = [];
                struct.push([keysA[i], nextStruct]);
                if (!isObjectInitialTempletable(a[keysA[i]], b[keysB[i]], nextStruct, level + 1)) {
                    return false;
                }
                else {
                    hasNested = true;
                }
            }
            else {
                struct.push([keysA[i]]);
            }
        }
        return level > 0 || hasNested;
    }
    function isObjectTempletable(obj, struct) {
        if (!obj) {
            return false;
        }
        for (let i = 0; i < struct.length; i++) {
            const key = struct[i][0];
            const isNested = struct[i].length > 1;
            if (obj[key] === void 0) {
                return false;
            }
            if (isNested) {
                if (!isObjectTempletable(obj[key], struct[i][1])) {
                    return false;
                }
            }
            else {
                if (isObject(obj[key])) {
                    return false;
                }
            }
        }
        return true;
    }
    function packNestedObjectTemplate(context, struct, invertedIndex, writer, options) {
        writer.write(constants_1.NESTED_TEMPLATE_OBJECT_START);
        for (let i = 0; i < struct.length; i++) {
            const key = struct[i][0];
            const isNested = struct[i].length > 1;
            packString(context, key, invertedIndex, writer, options);
            if (isNested) {
                packNestedObjectTemplate(context, struct[i][1], invertedIndex, writer, options);
            }
        }
        ;
        writer.write(constants_1.NESTED_TEMPLATE_OBJECT_END);
    }
    function packNestedObjectTemplateValues(context, struct, obj, invertedIndex, writer, options) {
        for (let i = 0; i < struct.length; i++) {
            const key = struct[i][0];
            const value = obj[key];
            const isNested = struct[i].length > 1;
            if (isNested) {
                packNestedObjectTemplateValues(context, struct[i][1], value, invertedIndex, writer, options);
            }
            else {
                packAny(context, value, invertedIndex, writer, options);
            }
        }
        ;
    }
    function packObject(context, obj, invertedIndex, writer, options) {
        writer.write(constants_1.OBJECT_START_TOKEN);
        const keys = Object.keys(obj);
        let templating = false;
        let templateObject;
        let templateStruct = [];
        if (keys.length > 1 && isObjectInitialTempletable(obj[keys[0]], obj[keys[1]], templateStruct)) {
            templateObject = obj[keys[0]];
            templating = true;
        }
        for (let i = 0; i < keys.length; i++) {
            if (templating && i > 0 && !isObjectTempletable(obj[keys[i]], templateStruct)) {
                templating = false;
                writer.write(constants_1.NESTED_TEMPLATE_OBJECT_FINAL);
            }
            if (templating) {
                if (i === 0) {
                    packNestedObjectTemplate(context, templateStruct, invertedIndex, writer, options);
                }
                packString(context, keys[i], invertedIndex, writer, options);
                packNestedObjectTemplateValues(context, templateStruct, obj[keys[i]], invertedIndex, writer, options);
            }
            else {
                const key = keys[i];
                const val = obj[key];
                if (val !== undefined) {
                    packString(context, key, invertedIndex, writer, options);
                    packAny(context, val, invertedIndex, writer, options);
                }
            }
        }
        ;
        if (templating) {
            writer.write(constants_1.NESTED_TEMPLATE_OBJECT_FINAL);
        }
        writer.write(constants_1.OBJECT_END_TOKEN);
    }
    function packObjectTemplate(context, obj, invertedIndex, writer, options) {
        writer.write(constants_1.TEMPLATE_OBJECT_START);
        const keys = Object.keys(obj);
        for (let i = 0; i < keys.length; i++) {
            const key = keys[i];
            const val = obj[key];
            if (val !== undefined) {
                packString(context, key, invertedIndex, writer, options);
            }
        }
        ;
        writer.write(constants_1.TEMPLATE_OBJECT_END);
    }
    function packObjectTemplateValues(context, obj, invertedIndex, writer, options) {
        const keys = Object.keys(obj);
        for (let i = 0; i < keys.length; i++) {
            const key = keys[i];
            const val = obj[key];
            if (val !== undefined) {
                packAny(context, val, invertedIndex, writer, options);
            }
        }
        ;
    }
    function isTemplatable(a, b) {
        if (a === undefined || b === undefined) {
            return false;
        }
        const keysA = Object.keys(a);
        if (keysA.length < 1) {
            return false;
        }
        const keysB = Object.keys(b);
        if (keysB.length < 1) {
            return false;
        }
        if (keysA.length !== keysB.length) {
            return false;
        }
        for (let i = 0; i < keysA.length; i++) {
            if (keysA[i] !== keysB[i]) {
                return false;
            }
        }
        return true;
    }
    function packArray(context, array, invertedIndex, writer, options) {
        context.arrayLevel++;
        if (context.arrayLevel > context.arrayItemWriters.length) {
            context.arrayItemWriters.push(new pack_writer_1.ZipsonStringPackWriter());
        }
        const arrayItemWriter = context.arrayItemWriters[context.arrayLevel - 1];
        const parentWriter = context.arrayItemWriters[context.arrayLevel - 2] || writer;
        parentWriter.write(constants_1.ARRAY_START_TOKEN);
        let previousItem = '';
        let repeatedTimes = 0;
        let repeatManyCount = 0;
        let templating = false;
        let templateObject;
        if (array.length > 1 && isObject(array[0]) && isObject(array[1])) {
            if (isTemplatable(array[0], array[1])) {
                templateObject = array[0];
                templating = true;
            }
        }
        for (let i = 0; i < array.length; i++) {
            let item = array[i];
            arrayItemWriter.value = '';
            if (item === undefined) {
                item = null;
            }
            if (templating && i > 0 && !isTemplatable(templateObject, item)) {
                templating = false;
                arrayItemWriter.write(constants_1.TEMPLATE_OBJECT_FINAL);
            }
            if (templating) {
                if (i === 0) {
                    packObjectTemplate(context, item, invertedIndex, arrayItemWriter, options);
                }
                packObjectTemplateValues(context, item, invertedIndex, arrayItemWriter, options);
            }
            else {
                packAny(context, item, invertedIndex, arrayItemWriter, options);
            }
            if (arrayItemWriter.value === previousItem) {
                repeatedTimes++;
                if (repeatedTimes >= constants_1.ARRAY_REPEAT_COUNT_THRESHOLD) {
                    if (repeatManyCount === 0) {
                        parentWriter.write(constants_1.ARRAY_REPEAT_MANY_TOKEN);
                        repeatManyCount++;
                    }
                    repeatManyCount++;
                }
                else {
                    parentWriter.write(constants_1.ARRAY_REPEAT_TOKEN);
                }
            }
            else {
                repeatedTimes = 0;
                if (repeatManyCount > 0) {
                    // write that repeated count first
                    parentWriter.write(util_1.compressInteger(repeatManyCount));
                    repeatManyCount = 0;
                }
                parentWriter.write(arrayItemWriter.value);
                previousItem = arrayItemWriter.value;
            }
        }
        if (repeatManyCount > 0) {
            parentWriter.write(util_1.compressInteger(repeatManyCount));
        }
        if (templating) {
            parentWriter.write(constants_1.TEMPLATE_OBJECT_FINAL);
        }
        parentWriter.write(constants_1.ARRAY_END_TOKEN);
        context.arrayLevel--;
    }
    function packNumber(context, obj, invertedIndex, writer, options) {
        let foundRef;
        if (obj % 1 === 0) {
            if (obj < 10 && obj > -10) {
                if (obj >= 0) {
                    writer.write(`${constants_1.INTEGER_SMALL_POS_TOKEN[obj]}`);
                }
                else {
                    writer.write(`${constants_1.INTEGER_SMALL_NEG_TOKEN[-obj - 1]}`);
                }
            }
            else if ((foundRef = invertedIndex.integerMap.get(obj)) != null) {
                writer.write(`${constants_1.REF_INTEGER_TOKEN}${foundRef}`);
            }
            else {
                const ref = util_1.compressInteger(invertedIndex.integerMap.size);
                const compressedInteger = util_1.compressInteger(obj);
                const newRef = `${constants_1.INTEGER_TOKEN}${compressedInteger}`;
                if (ref.length + constants_1.REFERENCE_HEADER_LENGTH < newRef.length) {
                    invertedIndex.integerMap.set(obj, ref);
                    writer.write(newRef);
                }
                else {
                    writer.write(`${constants_1.UNREFERENCED_INTEGER_TOKEN}${compressedInteger}`);
                }
            }
        }
        else {
            // Compress float prior to lookup to reuse for "same" floating values
            const compressedFloat = util_1.compressFloat(obj);
            if ((foundRef = invertedIndex.floatMap.get(compressedFloat))) {
                writer.write(`${constants_1.REF_FLOAT_TOKEN}${foundRef}`);
            }
            else {
                const ref = util_1.compressInteger(invertedIndex.floatMap.size);
                const newRef = `${constants_1.FLOAT_TOKEN}${compressedFloat}`;
                if (ref.length + constants_1.REFERENCE_HEADER_LENGTH < newRef.length) {
                    invertedIndex.floatMap.set(compressedFloat, ref);
                    writer.write(newRef);
                }
                else {
                    writer.write(`${constants_1.UNREFERENCED_FLOAT_TOKEN}${compressedFloat}`);
                }
            }
        }
    }
    function packString(context, obj, invertedIndex, writer, options) {
        let foundRef;
        if (options.detectUtcTimestamps && obj[obj.length - 1] === 'Z' && obj.match(constants_1.DATE_REGEX)) {
            const date = Date.parse(obj);
            packDate(context, date, invertedIndex, writer, options);
        }
        else if ((foundRef = invertedIndex.stringMap.get(obj))) {
            writer.write(`${constants_1.REF_STRING_TOKEN}${foundRef}`);
        }
        else {
            const ref = util_1.compressInteger(invertedIndex.stringMap.size);
            const newRef = `${constants_1.STRING_TOKEN}${obj.replace(constants_1.REGEX_STRING_TOKEN, constants_1.ESCAPED_STRING_TOKEN)}${constants_1.STRING_TOKEN}`;
            if (ref.length + constants_1.REFERENCE_HEADER_LENGTH + 1 < newRef.length) {
                invertedIndex.stringMap.set(obj, ref);
                writer.write(newRef);
            }
            else {
                writer.write(`${constants_1.UNREFERENCED_STRING_TOKEN}${obj.replace(constants_1.REGEX_UNREFERENCED_STRING_TOKEN, constants_1.ESCAPED_UNREFERENCED_STRING_TOKEN)}${constants_1.UNREFERENCED_STRING_TOKEN}`);
            }
        }
    }
    function packDate(context, obj, invertedIndex, writer, options) {
        let foundRef;
        const lowPrecisionDate = obj / constants_1.DATE_LOW_PRECISION;
        const isLowPrecision = lowPrecisionDate % 1 === 0;
        if (isLowPrecision) {
            if ((foundRef = invertedIndex.lpDateMap.get(lowPrecisionDate))) {
                writer.write(`${constants_1.REF_LP_DATE_TOKEN}${foundRef}`);
            }
            else {
                const ref = util_1.compressInteger(invertedIndex.lpDateMap.size);
                const compressedDate = util_1.compressInteger(lowPrecisionDate);
                const newRef = `${constants_1.LP_DATE_TOKEN}${compressedDate}`;
                if (ref.length + constants_1.REFERENCE_HEADER_LENGTH + 1 < newRef.length) {
                    invertedIndex.lpDateMap.set(lowPrecisionDate, ref);
                    writer.write(newRef);
                }
                else {
                    writer.write(`${constants_1.UNREFERENCED_LP_DATE_TOKEN}${compressedDate}`);
                }
            }
        }
        else {
            if ((foundRef = invertedIndex.dateMap.get(obj))) {
                writer.write(`${constants_1.REF_DATE_TOKEN}${foundRef}`);
            }
            else {
                const ref = util_1.compressInteger(invertedIndex.dateMap.size);
                const compressedDate = util_1.compressInteger(obj);
                const newRef = `${constants_1.DATE_TOKEN}${compressedDate}`;
                if (ref.length + constants_1.REFERENCE_HEADER_LENGTH + 1 < newRef.length) {
                    invertedIndex.dateMap.set(obj, ref);
                    writer.write(newRef);
                }
                else {
                    writer.write(`${constants_1.UNREFERENCED_DATE_TOKEN}${compressedDate}`);
                }
            }
        }
    }
    function isObject(obj) {
        const type = typeof obj;
        if (type === 'number') {
            return false;
        }
        else if (type === 'string') {
            return false;
        }
        else if (type === 'boolean') {
            return false;
        }
        else if (obj === null) {
            return false;
        }
        else if (Array.isArray(obj)) {
            return false;
        }
        else if (obj instanceof Date) {
            return false;
        }
        else {
            return true;
        }
    }
    function packAny(context, obj, invertedIndex, writer, options) {
        const type = typeof obj;
        if (type === 'number') {
            packNumber(context, obj, invertedIndex, writer, options);
        }
        else if (type === 'string') {
            packString(context, obj, invertedIndex, writer, options);
        }
        else if (type === 'boolean') {
            writer.write(obj ? constants_1.BOOLEAN_TRUE_TOKEN : constants_1.BOOLEAN_FALSE_TOKEN);
        }
        else if (obj === null) {
            writer.write(constants_1.NULL_TOKEN);
        }
        else if (obj === undefined) {
            writer.write(constants_1.UNDEFINED_TOKEN);
        }
        else if (Array.isArray(obj)) {
            packArray(context, obj, invertedIndex, writer, options);
        }
        else if (obj instanceof Date) {
            packDate(context, obj.getTime(), invertedIndex, writer, options);
        }
        else {
            packObject(context, obj, invertedIndex, writer, options);
        }
    }
    exports.packAny = packAny;
    function makePackContext() {
        return {
            arrayItemWriters: [],
            arrayLevel: 0,
        };
    }
    exports.makePackContext = makePackContext;
    function makeInvertedIndex() {
        return {
            stringMap: new Map(),
            integerMap: new Map(),
            floatMap: new Map(),
            dateMap: new Map(),
            lpDateMap: new Map(),
        };
    }
    exports.makeInvertedIndex = makeInvertedIndex;
});
//# sourceMappingURL=pack.js.map