(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "./util", "./constants"], factory);
    }
})(function (require, exports) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    const util_1 = require("./util");
    const constants_1 = require("./constants");
    var TargetType;
    (function (TargetType) {
        TargetType[TargetType["ARRAY"] = 0] = "ARRAY";
        TargetType[TargetType["OBJECT"] = 1] = "OBJECT";
        TargetType[TargetType["SCALAR"] = 2] = "SCALAR";
    })(TargetType || (TargetType = {}));
    const SKIP_SCALAR = {};
    function decompressScalar(token, data, cursor, orderedIndex) {
        const startIndex = cursor.index;
        let endIndex = cursor.index + 1;
        // Find end index of token value
        let foundStringToken;
        if (((token === constants_1.STRING_TOKEN) && (foundStringToken = constants_1.STRING_TOKEN))
            || ((token === constants_1.UNREFERENCED_STRING_TOKEN) && (foundStringToken = constants_1.UNREFERENCED_STRING_TOKEN))) {
            let escaped = true;
            while (escaped && endIndex < data.length) {
                endIndex = data.indexOf(foundStringToken, endIndex);
                escaped = data[endIndex - 1] === constants_1.ESCAPE_CHARACTER;
                endIndex++;
            }
            if (endIndex < startIndex) {
                endIndex = data.length;
            }
        }
        else {
            while (!(data.charCodeAt(endIndex) > constants_1.DELIMITING_TOKENS_THRESHOLD) && endIndex < data.length) {
                endIndex++;
            }
        }
        if (!cursor.drain && endIndex === data.length) {
            return SKIP_SCALAR;
        }
        // Update cursor end index
        cursor.index = endIndex - 1;
        // Decompress the token value
        let foundSmallInt;
        if ((foundSmallInt = constants_1.INTEGER_SMALL_TOKENS.get(token)) != null) {
            return foundSmallInt;
        }
        else if (token === constants_1.ARRAY_REPEAT_MANY_TOKEN) {
            return util_1.uncompressInteger(data.substring(startIndex + 1, endIndex));
        }
        else if (token === constants_1.REF_STRING_TOKEN) {
            return orderedIndex.strings[util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))];
        }
        else if (token === constants_1.REF_INTEGER_TOKEN) {
            return orderedIndex.integers[util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))];
        }
        else if (token === constants_1.REF_FLOAT_TOKEN) {
            return orderedIndex.floats[util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))];
        }
        else if (token === constants_1.REF_DATE_TOKEN) {
            return orderedIndex.dates[util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))];
        }
        else if (token === constants_1.REF_LP_DATE_TOKEN) {
            return orderedIndex.lpDates[util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))];
        }
        else if (token === constants_1.STRING_TOKEN) {
            return orderedIndex.strings[orderedIndex.strings.length] = data.substring(startIndex + 1, endIndex - 1).replace(constants_1.REGEX_ESCAPED_STRING_TOKEN, constants_1.STRING_TOKEN);
        }
        else if (token === constants_1.INTEGER_TOKEN) {
            return orderedIndex.integers[orderedIndex.integers.length] = util_1.uncompressInteger(data.substring(startIndex + 1, endIndex));
        }
        else if (token === constants_1.FLOAT_TOKEN) {
            return orderedIndex.floats[orderedIndex.floats.length] = util_1.uncompressFloat(data.substring(startIndex + 1, endIndex));
        }
        else if (token === constants_1.DATE_TOKEN) {
            return orderedIndex.dates[orderedIndex.dates.length] = new Date(util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))).toISOString();
        }
        else if (token === constants_1.LP_DATE_TOKEN) {
            return orderedIndex.lpDates[orderedIndex.lpDates.length] = new Date(constants_1.DATE_LOW_PRECISION * util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))).toISOString();
        }
        else if (token === constants_1.UNREFERENCED_STRING_TOKEN) {
            return data.substring(startIndex + 1, endIndex - 1).replace(constants_1.REGEX_UNREFERENCED_ESCAPED_STRING_TOKEN, constants_1.UNREFERENCED_STRING_TOKEN);
        }
        else if (token === constants_1.UNREFERENCED_INTEGER_TOKEN) {
            return util_1.uncompressInteger(data.substring(startIndex + 1, endIndex));
        }
        else if (token === constants_1.UNREFERENCED_FLOAT_TOKEN) {
            return util_1.uncompressFloat(data.substring(startIndex + 1, endIndex));
        }
        else if (token === constants_1.UNREFERENCED_DATE_TOKEN) {
            return new Date(util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))).toISOString();
        }
        else if (token === constants_1.UNREFERENCED_LP_DATE_TOKEN) {
            return new Date(constants_1.DATE_LOW_PRECISION * util_1.uncompressInteger(data.substring(startIndex + 1, endIndex))).toISOString();
        }
        else if (token === constants_1.BOOLEAN_TRUE_TOKEN) {
            return true;
        }
        else if (token === constants_1.BOOLEAN_FALSE_TOKEN) {
            return false;
        }
        else if (token === constants_1.NULL_TOKEN) {
            return null;
        }
        else if (token === constants_1.UNDEFINED_TOKEN) {
            return undefined;
        }
        throw new Error(`Unexpected scalar ${token} at ${startIndex}-${endIndex}`);
    }
    exports.decompressScalar = decompressScalar;
    function decompressPhase(cursor, data, orderedIndex) {
        for (; cursor.index < data.length; cursor.index++) {
            const c = data[cursor.index];
            if (c === constants_1.ARRAY_START_TOKEN) {
                cursor.currentTarget = { type: TargetType.ARRAY, value: [] };
                cursor.stack[++cursor.pointer] = cursor.currentTarget;
            }
            else if (c === constants_1.OBJECT_START_TOKEN) {
                cursor.currentTarget = { type: TargetType.OBJECT, value: {} };
                cursor.stack[++cursor.pointer] = cursor.currentTarget;
            }
            else if (c === constants_1.ARRAY_REPEAT_TOKEN) {
                const repeatedItem = cursor.currentTarget.value[cursor.currentTarget.value.length - 1];
                if (data[cursor.index + 1] === constants_1.ARRAY_REPEAT_MANY_TOKEN) {
                    cursor.index++;
                    const repeatCount = decompressScalar(data[cursor.index], data, cursor, orderedIndex);
                    for (let i = 0; i < repeatCount; i++) {
                        cursor.currentTarget.value.push(repeatedItem);
                    }
                }
                else {
                    cursor.currentTarget.value.push(repeatedItem);
                }
            }
            else {
                let targetValue;
                let currentTarget = cursor.currentTarget;
                let currentTargetType = currentTarget.type;
                if (c === constants_1.ARRAY_END_TOKEN || c === constants_1.OBJECT_END_TOKEN) {
                    targetValue = currentTarget.value;
                    currentTarget = cursor.currentTarget = cursor.stack[cursor.pointer - 1];
                    currentTargetType = currentTarget.type;
                    cursor.pointer--;
                }
                else {
                    targetValue = decompressScalar(c, data, cursor, orderedIndex);
                    if (targetValue === SKIP_SCALAR) {
                        return;
                    }
                }
                if (currentTargetType === TargetType.SCALAR) {
                    currentTarget.value = targetValue;
                }
                else if (currentTargetType === TargetType.ARRAY) {
                    currentTarget.value[currentTarget.value.length] = targetValue;
                }
                else if (currentTargetType === TargetType.OBJECT) {
                    if (currentTarget.key) {
                        currentTarget.value[currentTarget.key] = targetValue;
                        currentTarget.key = void 0;
                    }
                    else {
                        currentTarget.key = targetValue;
                    }
                }
            }
        }
    }
    exports.decompressPhase = decompressPhase;
    function decompress(data, orderedIndex) {
        const rootTarget = { type: TargetType.SCALAR, value: void 0 };
        const stack = new Array(10);
        stack[0] = rootTarget;
        const cursor = { index: 0, stack, currentTarget: rootTarget, pointer: 0, drain: true };
        decompressPhase(cursor, data, orderedIndex);
        return rootTarget.value;
    }
    exports.decompress = decompress;
    function decompressIncremental(orderedIndex) {
        const rootTarget = { type: TargetType.SCALAR, value: void 0 };
        const stack = new Array(10);
        stack[0] = rootTarget;
        const cursor = { index: 0, stack, currentTarget: rootTarget, pointer: 0, drain: false };
        let buffer = '';
        function increment(data) {
            if (data === null) {
                cursor.drain = true;
            }
            else if (data.length === 0) {
                return;
            }
            else {
                buffer += data;
            }
            console.log('buffer size', buffer.length);
            const cursorIndexBefore = cursor.index;
            decompressPhase(cursor, buffer, orderedIndex);
            const movedAmount = cursor.index - cursorIndexBefore;
            if (movedAmount > 0) {
                buffer = buffer.substring(movedAmount);
                cursor.index -= movedAmount;
            }
        }
        return { increment, cursor };
    }
    exports.decompressIncremental = decompressIncremental;
    function makeOrderedIndex() {
        // TODO: Improve this stuff somewhat, like preallocating or something
        return {
            strings: [],
            integers: [],
            floats: [],
            dates: [],
            lpDates: [],
        };
    }
    exports.makeOrderedIndex = makeOrderedIndex;
});
//# sourceMappingURL=unpack.js.map