(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "./constants", "./util"], factory);
    }
})(function (require, exports) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    const C = require("./constants");
    const util_1 = require("./util");
    function unpackScalar(data, orderedIndex) {
        if (data.length === 0) {
            return void 0;
        }
        else if (data[0] === C.REF_STRING_TOKEN) {
            return orderedIndex.strings[util_1.uncompressInteger(data.slice(1))];
        }
        else if (data[0] === C.REF_INTEGER_TOKEN) {
            return orderedIndex.integers[util_1.uncompressInteger(data.slice(1))];
        }
        else if (data[0] === C.REF_FLOAT_TOKEN) {
            return orderedIndex.floats[util_1.uncompressInteger(data.slice(1))];
        }
        else if (data[0] === C.REF_DATE_TOKEN) {
            return orderedIndex.dates[util_1.uncompressInteger(data.slice(1))];
        }
        else if (data[0] === C.REF_LP_DATE_TOKEN) {
            return orderedIndex.lpDates[util_1.uncompressInteger(data.slice(1))];
        }
        else if (data[0] === C.STRING_TOKEN) {
            return orderedIndex.strings[orderedIndex.strings.length] = data.slice(1, data.length - 1).replace(C.REGEX_ESCAPED_STRING_TOKEN, C.STRING_TOKEN);
        }
        else if (data[0] === C.INTEGER_TOKEN) {
            return orderedIndex.integers[orderedIndex.integers.length] = util_1.uncompressInteger(data.slice(1));
        }
        else if (data[0] === C.FLOAT_TOKEN) {
            return orderedIndex.floats[orderedIndex.floats.length] = util_1.uncompressFloat(data.slice(1));
        }
        else if (data[0] === C.DATE_TOKEN) {
            return orderedIndex.dates[orderedIndex.dates.length] = new Date(util_1.uncompressInteger(data.slice(1))).toISOString();
        }
        else if (data[0] === C.LP_DATE_TOKEN) {
            return orderedIndex.lpDates[orderedIndex.lpDates.length] = new Date(C.DATE_LOW_PRECISION * util_1.uncompressInteger(data.slice(1))).toISOString();
        }
        else if (data[0] === C.UNREFERENCED_STRING_TOKEN) {
            return data.slice(1, data.length - 1).replace(C.REGEX_UNREFERENCED_ESCAPED_STRING_TOKEN, C.UNREFERENCED_STRING_TOKEN);
        }
        else if (data[0] === C.UNREFERENCED_INTEGER_TOKEN) {
            return util_1.uncompressInteger(data.slice(1));
        }
        else if (data[0] === C.UNREFERENCED_FLOAT_TOKEN) {
            return util_1.uncompressFloat(data.slice(1));
        }
        else if (data[0] === C.UNREFERENCED_DATE_TOKEN) {
            return new Date(util_1.uncompressInteger(data.slice(1))).toISOString();
        }
        else if (data[0] === C.UNREFERENCED_LP_DATE_TOKEN) {
            return new Date(C.DATE_LOW_PRECISION * util_1.uncompressInteger(data.slice(1))).toISOString();
        }
        else if (data[0] === C.BOOLEAN_TRUE_TOKEN) {
            return true;
        }
        else if (data[0] === C.BOOLEAN_FALSE_TOKEN) {
            return false;
        }
        else if (data[0] === C.NULL_TOKEN) {
            return null;
        }
        else if (data[0] === C.UNDEFINED_TOKEN) {
            return undefined;
        }
        else {
            let smallInt;
            if ((smallInt = C.INTEGER_SMALL_POS_TOKEN.indexOf(data[0])) > -1) {
                return smallInt;
            }
            else if ((smallInt = C.INTEGER_SMALL_NEG_TOKEN.indexOf(data[0])) > -1) {
                return -smallInt - 1;
            }
        }
        throw new Error(`unknown scalar type ${data}`);
    }
    function unpackArray(data, orderedIndex) {
        const array = [];
        let lastItemIndex = 1;
        let previousItem = '';
        let level = 0;
        let i = 1;
        for (; i < data.length - 1; i++) {
            // TODO: Instead do a regex search to find the next delimiter
            //       withing that range count the number of array start and array end tokens and add to level
            //       then move to next
            if (level === 0 && C.DELIMITING_TOKENS.has(data[i])) {
                if (i > lastItemIndex) {
                    const itemSlice = data.slice(lastItemIndex, i);
                    if (itemSlice[0] === C.ARRAY_REPEAT_TOKEN) {
                        if (itemSlice[1] === C.ARRAY_REPEAT_MANY_TOKEN) {
                            const repeatManyCount = util_1.uncompressInteger(itemSlice.slice(2));
                            for (let x = 0; x < repeatManyCount; x++) {
                                array.push(previousItem);
                            }
                        }
                        else {
                            array.push(previousItem);
                        }
                    }
                    else {
                        previousItem = unpackAny(itemSlice, orderedIndex);
                        array.push(previousItem);
                    }
                }
                lastItemIndex = i;
            }
            if (data[i] === C.ARRAY_START_TOKEN || data[i] === C.OBJECT_START_TOKEN) {
                level++;
            }
            else if (data[i] === C.ARRAY_END_TOKEN || data[i] === C.OBJECT_END_TOKEN) {
                level--;
            }
            else if (data[i] === C.STRING_TOKEN) {
                let escaped = true;
                while (escaped) {
                    i = data.indexOf(C.STRING_TOKEN, i + 1);
                    escaped = data[i - 1] === C.ESCAPE_CHARACTER;
                }
                if (i < 0) {
                    i = data.length - 1;
                }
            }
            else if (data[i] === C.UNREFERENCED_STRING_TOKEN) {
                let escaped = true;
                while (escaped) {
                    i = data.indexOf(C.UNREFERENCED_STRING_TOKEN, i + 1);
                    escaped = data[i - 1] === C.ESCAPE_CHARACTER;
                }
                if (i < 0) {
                    i = data.length - 1;
                }
            }
        }
        if (i > lastItemIndex && data[i] === C.ARRAY_END_TOKEN) {
            const itemSlice = data.slice(lastItemIndex, i);
            if (itemSlice[0] === C.ARRAY_REPEAT_TOKEN) {
                if (itemSlice[1] === C.ARRAY_REPEAT_MANY_TOKEN) {
                    const repeatManyCount = util_1.uncompressInteger(itemSlice.slice(2));
                    for (let x = 0; x < repeatManyCount; x++) {
                        array.push(previousItem);
                    }
                }
                else {
                    array.push(previousItem);
                }
            }
            else {
                previousItem = unpackAny(itemSlice, orderedIndex);
                array.push(previousItem);
            }
        }
        return array;
    }
    function unpackObject(data, orderedIndex) {
        const object = {};
        let lastBufferIndex = 1;
        let lastItemIndex = 1;
        let level = 0;
        let i = 1;
        let itemBuffer = [];
        for (; i < data.length - 1; i++) {
            if (level === 0 && C.DELIMITING_TOKENS.has(data[i])) {
                if (i > lastItemIndex) {
                    itemBuffer.push(data.slice(lastItemIndex, i));
                    if (itemBuffer.length === 2) {
                        object[unpackScalar(itemBuffer[0], orderedIndex)] = unpackAny(itemBuffer[1], orderedIndex);
                        itemBuffer.length = 0;
                        lastBufferIndex = i;
                    }
                }
                lastItemIndex = i;
            }
            if (data[i] === C.ARRAY_START_TOKEN || data[i] === C.OBJECT_START_TOKEN) {
                level++;
            }
            else if (data[i] === C.ARRAY_END_TOKEN || data[i] === C.OBJECT_END_TOKEN) {
                level--;
            }
            else if (data[i] === C.STRING_TOKEN) {
                let escaped = true;
                while (escaped) {
                    i = data.indexOf(C.STRING_TOKEN, i + 1);
                    escaped = data[i - 1] === C.ESCAPE_CHARACTER;
                }
                if (i < 0) {
                    i = data.length - 1;
                }
            }
            else if (data[i] === C.UNREFERENCED_STRING_TOKEN) {
                let escaped = true;
                while (escaped) {
                    i = data.indexOf(C.UNREFERENCED_STRING_TOKEN, i + 1);
                    escaped = data[i - 1] === C.ESCAPE_CHARACTER;
                }
                if (i < 0) {
                    i = data.length - 1;
                }
            }
        }
        if (i > lastItemIndex && data[i] === C.OBJECT_END_TOKEN) {
            itemBuffer.push(data.slice(lastItemIndex, i));
            if (itemBuffer.length === 2) {
                if (itemBuffer.length === 2) {
                    object[unpackScalar(itemBuffer[0], orderedIndex)] = unpackAny(itemBuffer[1], orderedIndex);
                    itemBuffer.length = 0;
                }
                itemBuffer.length = 0;
            }
        }
        return object;
    }
    function unpackAny(data, orderedIndex) {
        if (data.length === 0) {
            return void 0;
        }
        else if (data[0] === C.OBJECT_START_TOKEN) {
            return unpackObject(data, orderedIndex);
        }
        else if (data[0] === C.ARRAY_START_TOKEN) {
            return unpackArray(data, orderedIndex);
        }
        else {
            return unpackScalar(data, orderedIndex);
        }
    }
    exports.unpackAny = unpackAny;
    function makeOrderedIndex() {
        return {
            strings: [],
            integers: [],
            floats: [],
            dates: [],
            lpDates: [],
        };
    }
    exports.makeOrderedIndex = makeOrderedIndex;
});
//# sourceMappingURL=unpackold.js.map